#include <stdint.h>
#include <hls_stream.h>
#include "ap_int.h"

#include "mpc.h"

float simulate_state(
                    const float sw[3],
					float Vdc,
					float io_k[2],
					float vref[2],
					float ii_k1[3],
					float vc_k1[3],
					uint32_t mode) {
	#pragma HLS inline

	float vi_k1[3];
	float ii_k2[3];
	float vc_k2[3];
	float iiref_k2[3];

	float cost = 0;

    // 2nd-step calculations for the alpha/beta-components
	for(int i=0;i<2;i++) {
		#pragma HLS pipeline II=1

		vi_k1[i] = Vdc*sw[i];


		ii_k2[i] = Ad[0][0]*ii_k1[i] + Ad[0][1]*vc_k1[i] + Bd[0][0]*vi_k1[i] + Bd[0][1]*io_k[i];
		vc_k2[i] = Ad[1][0]*ii_k1[i] + Ad[1][1]*vc_k1[i] + Bd[1][0]*vi_k1[i] + Bd[1][1]*io_k[i];


		iiref_k2[i] = (vref[i] - Ad[1][1]*vc_k2[i] - Bd[1][0]*vi_k1[i] - Bd[1][1]*io_k[i]) / Ad[1][0];

		cost += (iiref_k2[i]-ii_k2[i])*(iiref_k2[i]-ii_k2[i]);
	}

    // 2nd-step calculations for the 0-component
	if (mode == 1) {
		vi_k1[2] = Vdc*sw[2]-Vdc/2;

		ii_k2[2] = Ad0[0][0]*ii_k1[2] + Ad0[0][1]*vc_k1[2] + Bd0[0][0]*vi_k1[2] + Bd0[0][1]*io_k[2];

		cost += 50*ii_k2[2]*ii_k2[2];
	}

	return cost;
}

void mpc(	hls::stream<float>& in_Vdc,
			hls::stream<float>& in_iia,
			hls::stream<float>& in_iib,
			hls::stream<float>& in_ii0,
			hls::stream<float>& in_vca,
			hls::stream<float>& in_vcb,
			hls::stream<float>& in_vc0,
			hls::stream<float>& in_ioa,
			hls::stream<float>& in_iob,
			hls::stream<float>& in_io0,
			hls::stream<float>& in_vrefa,
			hls::stream<float>& in_vrefb,
			hls::stream<uint32_t>& in_mode,
			ap_uint<3>& out)
{

	// pragma commands
	#pragma HLS INTERFACE axis port=in_Vdc register_mode=both register
	#pragma HLS INTERFACE axis port=in_iia register_mode=both register
	#pragma HLS INTERFACE axis port=in_iib register_mode=both register
	#pragma HLS INTERFACE axis port=in_ii0 register_mode=both register
	#pragma HLS INTERFACE axis port=in_vca register_mode=both register
	#pragma HLS INTERFACE axis port=in_vcb register_mode=both register
	#pragma HLS INTERFACE axis port=in_vc0 register_mode=both register
	#pragma HLS INTERFACE axis port=in_ioa register_mode=both register
	#pragma HLS INTERFACE axis port=in_iob register_mode=both register
	#pragma HLS INTERFACE axis port=in_io0 register_mode=both register
	#pragma HLS INTERFACE axis port=in_vrefa register_mode=both register
	#pragma HLS INTERFACE axis port=in_vrefb register_mode=both register
	#pragma HLS INTERFACE axis port=in_mode register_mode=both register
	#pragma HLS INTERFACE mode=ap_vld port=out register
	#pragma HLS INTERFACE ap_ctrl_none port=return

	// variables
	float Vdc = in_Vdc.read();
	float ii_k[3] = {in_iia.read(), in_iib.read(), in_ii0.read()};
	float vc_k[3] = {in_vca.read(), in_vcb.read(), in_vc0.read()};
	float io_k[3] = {in_ioa.read(), in_iob.read(), in_io0.read()};
	float vref[2] = {in_vrefa.read(), in_vrefb.read()};
	uint32_t mode = in_mode.read();

    // possible switching states
	// with leg_a = LSB, leg_c = MSB
	const ap_uint<3> states[8] = {0b000, 0b001, 0b011, 0b010, 0b110, 0b100, 0b101, 0b111};

    // normalized alpha/beta/0-components at the output of the inverter
    // corresponding to the 8 switching states
	const float sw[8][3] =
		{{0, 0, 0}, {_2_3, 0, _1_3}, {_1_3, _sqrt3_3, _2_3}, {-_1_3, _sqrt3_3, _1_3}, {-_2_3, 0, _2_3}, {-_1_3, -_sqrt3_3, _1_3}, {_1_3, -_sqrt3_3, _2_3}, {0, 0, 1}};

	// current state
	static int x_k = 0;

	// start of 1st step
	float vi_k[3];
    float ii_k1[3];
	float vc_k1[3];

    // 1st-step calculations for the alpha/beta-components
	for(int i=0;i<2;i++) {
		#pragma HLS pipeline II=1

		vi_k[i] = Vdc*sw[x_k][i];
	}

	for(int i=0;i<2;i++) {
		#pragma HLS pipeline II=1

		ii_k1[i] = Ad[0][0]*ii_k[i] + Ad[0][1]*vc_k[i] + Bd[0][0]*vi_k[i] + Bd[0][1]*io_k[i];
		vc_k1[i] = Ad[1][0]*ii_k[i] + Ad[1][1]*vc_k[i] + Bd[1][0]*vi_k[i] + Bd[1][1]*io_k[i];

	}

    // 1st-step calculations for the 0-component
	if (mode == 1) {
		vi_k[2] = Vdc*sw[x_k][2]-Vdc/2;

		ii_k1[2] = Ad0[0][0]*ii_k[2] + Ad0[0][1]*vc_k[2] + Bd0[0][0]*vi_k[2] + Bd0[0][1]*io_k[2];
		vc_k1[2] = Ad0[1][0]*ii_k[2] + Ad0[1][1]*vc_k[2] + Bd0[1][0]*vi_k[2] + Bd0[1][1]*io_k[2];
	}

	// start of 2nd step
	float c;
	float min_c = 1e18;
	int min_ind = 0;

	// simulation of all possible states and identification of the optimal one
	for (int i=0; i<8; i++) {
		#pragma HLS pipeline II=1

		c = simulate_state(sw[i], Vdc, io_k, vref, ii_k1, vc_k1, mode);

		if (c < min_c) {
			min_c = c;
			min_ind = i;
		}
	}

	// updating state and out
	x_k = min_ind;
	out.write(states[x_k]);
}
